@@ -5,8 +5,10 @@ import numpy as np # handle arrays
 import pandas as pd # data wrangling
 import copy as cp
 import joblib as jl
+import textdistance as td
+import scipy.spatial.distance as sp_dist
 
-def saveToDf(history,save_to_file,n_cores=0):
+def saveToDf(history,save_to_file,n_cores=0,verbose=10, **kwargs):
     """Save status of model to dataframe, write to file location given.
 
     Creates a pandas Dataframe in long format with the given model history, with
@@ -30,6 +32,7 @@ def saveToDf(history,save_to_file,n_cores=0):
     Keyword arguments:
     n_cores -- number of cores to parallelize file export across, if 0, all
         cores available are used (default 0; int)
+    **kwargs -- additional arguents for joblib multiprocessing
 
     Returns:
     pandas dataframe with model history as described above
@@ -42,7 +45,8 @@ def saveToDf(history,save_to_file,n_cores=0):
 
     new_df = ','.join(
         ['Time','Population','Organism','ID','Pathogens','Protection','Alive']
-        ) + '\n' + '\n'.join( jl.Parallel(n_jobs=n_cores, verbose=10) (
+        ) + '\n' + '\n'.join( jl.Parallel(
+            n_jobs=n_cores, verbose=verbose, **kwargs) (
             jl.delayed( lambda d: ''.join(d) ) (
                 '\n'.join( [
                     '\n'.join( [ ','.join( [
@@ -150,9 +154,15 @@ def populationsDf(
         ] = compartment_names[2]
 
     grouped = grouped[ grouped['Compartment'] == compartment ]
+
+    grouped = grouped.groupby(
+        ['Time','Population','Compartment']
+    ).sum().reset_index()
+
     grouped = grouped.drop(
         columns=['Alive','Infected','Protected','Compartment']
         )
+
     grouped = grouped.pivot(
         columns='Population', values='Number', index='Time'
         ).fillna(0).reset_index('Time')
@@ -273,8 +283,9 @@ def compartmentDf(
 
 def compositionDf(
         data, populations=[], type_of_composition='Pathogens', hosts=True,
-        vectors=False, num_top_sequences=-1, track_specific_genomes=[],
-        save_to_file=""):
+        vectors=False, num_top_sequences=-1, track_specific_sequences=[],
+        genomic_positions=[], count_individuals_based_on_model=None,
+        save_to_file="", n_cores=0, **kwargs):
     """Create dataframe with counts for pathogen genomes or resistance.
 
     Creates a pandas Dataframe with dynamics of the pathogen strains or
@@ -299,11 +310,23 @@ def compositionDf(
     num_top_sequences -- how many sequences to count separately and include
         as columns, remainder will be counted under column "Other"; if <0,
         includes all genomes in model (default -1; int)
-    track_specific_genomes -- contains specific sequences to have
+    track_specific_sequences -- contains specific sequences to have
         as a separate column if not part of the top num_top_sequences
-        sequences (list of Strings)
+        sequences (default empty list; list of Strings)
+    genomic_positions -- list in which each element is a list with loci
+        positions to extract (e.g. genomic_positions=[ [0,3], [5,6] ] extracts
+        positions 0, 1, 2, and 5 from each genome); if empty, takes full genomes
+        (default empty list; list of lists of int)
+    count_individuals_based_on_model -- Model object with populations and
+        fitness functions used to evaluate the most fit pathogen genome in each
+        host/vector in order to count only a single pathogen per host/vector, as
+        opposed to all pathogens within each host/vector; if None, counts all
+        pathogens (default None; None or Model)
     save_to_file -- file path and name to save model data under, no saving
         occurs if empty string (default ''; String)
+    n_cores -- number of cores to parallelize processing across, if 0, all
+        cores available are used (default 0; int)
+    **kwargs -- additional arguents for joblib multiprocessing
 
     Returns:
     pandas dataframe with model sequence composition dynamics as described above
@@ -314,28 +337,69 @@ def compositionDf(
     else:
         dat = cp.deepcopy( data )
 
+    dat = dat[ (dat['Pathogens'] != "") & (~dat['Pathogens'].isna()) ]
+
+    if len(genomic_positions) > 0:
+        print('Extracting genomic locations...')
+        def extractSeq(ind):
+            seqs = [ p[ genomic_positions[0]:genomic_positions[1] ]
+                for p in ind['Pathogens'].split(';') ]
+
+            return ';'.join(seqs)
+
+        pathogens = np.array( jl.Parallel(n_jobs=n_cores, verbose=1, **kwargs) (
+            jl.delayed( extractSeq ) (ind) for ind in dat
+            ) )
+        dat['Pathogens'] = pathogens
+
+    if count_individuals_based_on_model is not None:
+        print('Collapsing infections to individuals...')
+        model = count_individuals_based_on_model
+        if type_of_composition != 'Pathogens':
+            raise ValueError("Computing count_individuals_based_on_model=True is only allowed for type_of_composition='Pathogens'.")
+
+        dat['patpop'] = dat['Pathogens'].fillna('') + ':::' + dat['Population']
+        unique_patpop = dat['patpop'].unique()
+
+        for i,combination in enumerate(unique_patpop):
+            print( str(i) + ' / ' + str(len(unique_patpop)) + ' combinations' )
+            gen_str,pop = combination.split(':::')
+            if gen_str != '':
+                genomes = gen_str.split(';')
+                values = np.array( [
+                    model.populations[pop].fitnessHost(p)
+                    for p in genomes ] )
+                dominant_pathogen = genomes[ np.argmax(values) ]
+                dat['patpop'] = dat['patpop'].replace(
+                    combination, dominant_pathogen
+                    )
+            else:
+                dat['patpop'] = dat['patpop'].replace( combination, '' )
+
+        dat['Pathogens'] = dat['patpop']
+
     if not hosts:
         dat = dat[ dat['Organism'] != 'Host' ]
     if not vectors:
         dat = dat[ dat['Organism'] != 'Vector' ]
 
-    if num_top_sequences < 0:
-        num_top_sequences = len( pd.unique( dat[type_of_composition] ) )
-
-    all_genomes = pd.Series(
+    all_sequences = pd.Series(
         ';'.join( dat[type_of_composition].dropna() ).split(';')
         ).str.strip()
-    top_genomes = all_genomes.value_counts(ascending=False)
+    top_sequences = all_sequences.value_counts(ascending=False)
 
-    if len(top_genomes) < num_top_sequences:
-        num_top_sequences = len(top_genomes)
+    if num_top_sequences < 0:
+        num_top_sequences = len( top_sequences )
 
-    genomes_to_track = list(top_genomes[0:num_top_sequences].index)
-    for genome in track_specific_genomes:
+    if len(top_sequences) < num_top_sequences:
+        num_top_sequences = len(top_sequences)
+
+    # genomes_to_track = list(top_sequences[0:num_top_sequences].index)
+    genomes_to_track = track_specific_sequences
+    for genome in list(top_sequences[0:num_top_sequences].index):
         if genome not in genomes_to_track:
             genomes_to_track.append(genome)
 
-
     genome_split_data = []
     other_genome_data = pd.DataFrame(
         np.zeros( len( pd.unique( data['Time'] ) ) ),
@@ -343,10 +407,10 @@ def compositionDf(
         )
     c = 0
 
-    if len( ''.join(top_genomes.index) ) > 0:
-        for genome in top_genomes.index:
+    if len( ''.join(top_sequences.index) ) > 0:
+        for genome in top_sequences.index:
             dat_genome = dat[ dat[type_of_composition].str.contains(
-                genome, na=False
+                genome, na=False, regex=False
                 ) ]
             grouped = dat_genome.groupby('Time').size().reset_index(name=genome)
             grouped = grouped.set_index('Time')
@@ -363,7 +427,7 @@ def compositionDf(
 
             c += 1
             print(
-                str(c) + ' / ' + str( len(top_genomes.index) )
+                str(c) + ' / ' + str( len(top_sequences.index) )
                 + ' genotypes processed.'
                 )
 
@@ -379,6 +443,19 @@ def compositionDf(
         )
     composition = times.join( genome_split_data, how='outer' )
     composition = composition.drop(columns=['*None*']).fillna(0)
+
+    new_order = []
+    for seq in track_specific_sequences:
+        new_order.append(seq)
+        if seq not in composition.columns:
+            composition[seq] = 0
+
+    for seq in composition.columns:
+        if seq not in new_order:
+            new_order.append(seq)
+
+    composition = composition[new_order]
+
     composition = composition.reset_index()
     composition.columns = ['Time'] + list( composition.columns )[1:]
 
@@ -408,7 +485,7 @@ def getPathogens(data, save_to_file=""):
     out = pd.Series( ';'.join(
         data['Pathogens'].dropna()
         ).split(';') ).str.strip().value_counts(ascending=False).reset_index()
-    out.columns = ['Pathogens'] + list( out.columns )[1:]
+    out.columns = ['Pathogens','Counts']
 
     if len(save_to_file) > 0:
         out.to_csv(save_to_file, index=False)
@@ -435,9 +512,187 @@ def getProtections(data, save_to_file=""):
     out = pd.Series( ';'.join(
         data['Protection'].dropna()
         ).split(';') ).str.strip().value_counts(ascending=False).reset_index()
-    out.columns = ['Protection'] + list( out.columns )[1:]
+    out.columns = ['Protection','Counts']
 
     if len(save_to_file) > 0:
         out.to_csv(save_to_file, index=False)
 
     return out
+
+def pathogenDistanceDf(
+        data, num_top_sequences=-1, track_specific_sequences=[], seq_names=[],
+        save_to_file="", n_cores=0):
+    """Create DataFrame with pairwise Hamming distances for pathogen sequences
+    in data.
+
+    DataFrame has indexes and columns named according to genomes or argument
+    seq_names, if passed. Distance is measured as percent Hamming distance from
+    an optimal genome sequence.
+
+    Arguments:
+    data -- dataframe with model history as produced by saveToDf function
+
+    Keyword arguments:
+    num_top_sequences -- how many sequences to include in matrix; if <0,
+        includes all genomes in data passed (default -1; int)
+    track_specific_sequences -- contains specific sequences to include in matrix
+        if not part of the top num_top_sequences sequences (default empty list;
+        list of Strings)
+    seq_names -- list with names to be used for sequence labels in matrix must
+        be of same length as number of sequences to be displayed; if empty,
+        uses sequences themselves (default empty list; list of Strings)
+    save_to_file -- file path and name to save model data under, no saving
+        occurs if empty string (default ''; String)
+    n_cores -- number of cores to parallelize distance compute across, if 0, all
+        cores available are used (default 0; int)
+
+    Returns:
+    pandas dataframe with distance matrix as described above
+    """
+
+    sequences = getPathogens(data)['Pathogens']
+
+    if num_top_sequences > 0:
+        sequences = sequences[0:num_top_sequences]
+        sequences = sequences.append(
+            pd.Series(track_specific_sequences)
+            ).unique()
+
+    # Fix — non parallelized
+    dis_mat = np.array([[td.hamming(s1, s2) / max(len(s1),1)
+        for s2 in sequences] for s1 in sequences])
+            # Added the max() to avoid division by zero error triggered when
+            # getPathogenDistanceHistoryDf calls this method with an empty
+            # dataframe (for a timepoint with no pathogens)
+
+    # For some reason, this code triggers a full rerun of the simulation.
+    # Possible joblib bug?
+    # if not n_cores:
+    #     n_cores = jl.cpu_count()
+    #
+    # dis_mat = np.array( jl.Parallel(n_jobs=n_cores, verbose=1) (
+    #     jl.delayed( lambda s1: [
+    #         td.hamming(s1, s2) / max(len(s1),1) for s2 in sequences
+    #         ] ) (s1) for s1 in sequences
+    #     ) )
+
+    names = sequences
+    if len(seq_names) > 0:
+        new_names = seq_names * int( np.ceil( len(names) / len(seq_names) ) )
+        names = new_names[0:len(names)]
+
+    dis_df = pd.DataFrame( dis_mat, index=names, columns=names )
+
+    if len(save_to_file) > 0:
+        dis_df.to_csv(save_to_file, index=True)
+
+    return dis_df
+
+def getPathogenDistanceHistoryDf(
+        data, samples=1, num_top_sequences=-1, track_specific_sequences=[],
+        seq_names=[], save_to_file="", n_cores=0):
+    """Create DataFrame with pairwise Hamming distances for pathogen sequences
+    in data.
+
+    DataFrame has indexes and columns named according to genomes or argument
+    seq_names, if passed. Distance is measured as percent Hamming distance from
+    an optimal genome sequence.
+
+    Arguments:
+    data -- dataframe with model history as produced by saveToDf function
+
+    Keyword arguments:
+    samples -- how many timepoints to uniformly sample from the total
+        timecourse; if <0, takes all timepoints (default 1; int)
+    num_top_sequences -- how many sequences to include in matrix; if <0,
+        includes all genomes in data passed (default -1; int)
+    track_specific_sequences -- contains specific sequences to include in matrix
+        if not part of the top num_top_sequences sequences (default empty list;
+        list of Strings)
+    seq_names -- list with names to be used for sequence labels in matrix must
+        be of same length as number of sequences to be displayed; if empty,
+        uses sequences themselves (default empty list; list of Strings)
+    save_to_file -- file path and name to save model data under, no saving
+        occurs if empty string (default ''; String)
+    n_cores -- number of cores to parallelize distance compute across, if 0, all
+        cores available are used (default 0; int)
+
+    Returns:
+    pandas dataframe with distance matrix as described above
+    """
+
+    if samples > 0:
+        samples = np.linspace(
+            0, len(pd.unique(data['Time']))-1, samples
+            ).astype(int)
+        sampled_times = pd.unique(data['Time'])[samples]
+        data = data[ data['Time'].isin(sampled_times) ]
+
+    grouped = data.groupby('Time')
+    dis_df = grouped.apply(
+        lambda d: pd.melt(
+            pathogenDistanceDf(
+                d, num_top_sequences=num_top_sequences,
+                    track_specific_sequences=track_specific_sequences,
+                    seq_names=seq_names, save_to_file="", n_cores=n_cores
+                    ).reset_index(),
+                id_vars=['Pathogens'], var_name='Pathogen_2',
+                value_name='Distance'
+                )
+        ).reset_index()
+
+    dis_df.columns = ['Time','drop','Pathogen_1','Pathogen_2','Distance']
+    dis_df = dis_df.drop(columns='drop')
+
+    if len(save_to_file) > 0:
+        dis_df.to_csv(save_to_file, index=False)
+
+    return dis_df
+
+def getGenomeTimesDf(
+        data, samples=1, save_to_file="", n_cores=0, **kwargs):
+    """Create DataFrame with times genomes first appeared during simulation.
+
+    Arguments:
+    data -- dataframe with model history as produced by saveToDf function
+
+    Keyword arguments:
+    samples -- how many timepoints to uniformly sample from the total
+        timecourse; if <0, takes all timepoints (default 1; int)
+    save_to_file -- file path and name to save model data under, no saving
+        occurs if empty string (default ''; String)
+    n_cores -- number of cores to parallelize across, if 0, all cores available
+        are used (default 0; int)
+    **kwargs -- additional arguents for joblib multiprocessing
+
+    Returns:
+    pandas dataframe with genomes and times as described above
+    """
+
+    if samples > 0:
+        samples = np.linspace(
+            0, len(pd.unique(data['Time']))-1, samples
+            ).astype(int)
+        sampled_times = pd.unique(data['Time'])[samples]
+        data = data[ data['Time'].isin(sampled_times) ]
+
+    his_dat = pd.DataFrame()
+
+    his_dat['Sequence'] = pd.Series(
+        ';'.join( dat['Pathogens'].dropna() ).split(';')
+        ).str.strip()
+
+    def getTime(seq):
+        t = his_dat['Sequence'].searchsorted(seq, side='left')
+        return t
+
+    his_dat['Time_emergence'] = jl.Parallel(
+        n_jobs=n_cores, verbose=1, **kwargs
+        ) (
+            jl.delayed( getTime ) (seq) for seq in his_dat['Sequence']
+            )
+
+    if len(save_to_file) > 0:
+        dis_df.to_csv(save_to_file, index=False)
+
+    return his_dat
